/*
Copyright (C) 2003 Martin A. Olevitch
    Written by Martin A. Olevitch (marty@cosray.wustl.edu)

    This file is part of Shmed.

    Shmed is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Shmed is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Shmed, see the file COPYING; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* shmed - simple shell form editor
 *
 *	USAGE:
 *	shmed [-pbc] [-f formfile] -o outfile
 *		-p		perl output	(default)
 *		-b		Bourne shell output
 *		-c		csh output
 *		-f formfile	name of file containing form (default stdin)
 *		-o outfile	name of output file (no default)
 *
 *	DIAGNOSTICS:
 *		shmed returns non-zero if there are startup
 *		problems or if it is interrupted, else 0.
 *
 *	FILES:
 *	    format of the form file
 *		Normal text lines are printed as they appear.
 *		Editable lines should begin with a '.' (dot), and
 *		contain 2 or 3 fields separated by a '~' (tilde).
 *			1st field: label to be printed in the editor.
 *			2nd field: name of variable to assign the value to
 *			3rd field: (optional) default value for the variable
 *
 *	    output file
 *		The output file contains a bunch of variable assignments
 *		in the language you selected (perl, sh, or csh). It should
 *		be require'd, .'d, or source'd after shmed is run.
 *
 *	AUTHOR
 *
 *		Marty Olevitch
 *		Physics Dept.
 *		Washington University
 *		St. Louis, MO 63130
 *
 *		marty@wuphys.wustl.edu
 *
 *		March, 1991
 */

#define VERSION "1.3"
#ifdef NOTDEF
/* revision history

1.6	4/14/04		Uses stdarg.h instead of varargs.h in util.c.

1.3	9/15/00		Used keypad() function so that arrow keys wouldn't
			make it malfunction. (MAO)

1.2	3/24/97		added function prototypes as returning voids, and
			removed "breaks" in switch that were causing warnings
			David Tristram - Tristram Visual - david@tristram.com

1.1	5/3/91		complain if we get a blank form file
			initialize 'done' (found by A. Proudfoot(foot@fluke.com))
			"support" for old curses

1.0	3/x/91		initial release to comp.sources.misc
*/
#endif

#include <curses.h>
#include <ctype.h>
#include <signal.h>

#define CTL(x)	((x) & 0x3f)
#define	ESC		0x1b
#define DEL		0x7f

#define	MAXSTR	500		/* For sizing strings -- DON'T use BUFSIZ! */
#define	STREQ(a, b)	(*(a) == *(b) && strcmp((a), (b)) == 0)
#define BEEP	putchar(CTL('G'))

/* CHOP - replace last character of a string */
#define CHOP(s, c)	((s)[strlen(s)-1] = (c))

#define FIELDSEP	"~"				/* form file field separator string */
#define PERLFMT		"$%s = '%s';\n"
#define BOURNEFMT	"%s='%s'\n"
#define CSHFMT		"set %s='%s'\n"

#define MAXFORMROWS	24				/* max. no. of rows in a form */
char answer[MAXFORMROWS][MAXSTR];	/* response lines */
int	editable[MAXFORMROWS];			/* tells if form line is editable */
char label[MAXFORMROWS][MAXSTR];	/* labels for editable lines */
char variable[MAXFORMROWS][MAXSTR];	/* variable names for editable lines */
int longval;						/* length of longest formrow string */
int nrows;							/* number of rows in our form */

WINDOW *helpwin;	/* help screen */
WINDOW *altwin;		/* for saving real screen when help is displayed */

char *progname;
FILE *efopen();

void sighandle();
void finish_up();
void startcurses();
void inithelp();
void help();

/*
 -	usage - print usage message
 */
void
usage()
{
	fprintf(stderr,"USAGE: %s [-pbc] [-f formfile] -o outfile\n",
		progname);
}

/*
 - main
 */
main(argc, argv)
	int argc;
	char *argv[];
{
	int bourneout = 0;			/* bourne shell output */
	int cshout = 0;				/* csh shell output */
	int done;
	char filename[MAXSTR];		/* form file */
	char fmt[MAXSTR];			/* printf format for output */
	int i;
	FILE *outfp;
	char outputfile[MAXSTR];	/* output file name */
	int outputflag = 0;			/* 1 if we have a valid -o argument */
	int perlout = 0;			/* perl output */

	int c;
	int errflg = 0;
	extern int optind;
	extern char *optarg;

	progname = argv[0];
	strcpy(filename, "-");
		
	while ((c = getopt(argc, argv, "pcbf:o:")) != EOF) {
		switch (c) {
		case 'p':	/* perl output */
			if(bourneout || cshout) {
				fprintf(stderr,"Use only one of -c, -b and -p flags\n");
				errflg++;
			} else 
				perlout = 1;
			break;
		case 'b':	/* bourne shell output */
			if(perlout || cshout)  {
				fprintf(stderr,"Use only one of -c, -b and -p flags\n");
				errflg++;
			} else 
				bourneout = 1;
			break;
		case 'c':	/* csh output */
			if(perlout || bourneout)  {
				fprintf(stderr,"Use only one of -c, -b and -p flags\n");
				errflg++;
			} else 
				cshout = 1;
			break;
		case 'f':	/* form file name */
			strcpy(filename, optarg);
			break;
		case 'o':	/* output file name */
			outputflag++;
			strcpy(outputfile, optarg);
			break;
		case '?':
		default:
			errflg++;
			break;
		}
	}
	
	if (errflg) {
		usage();
		exit(2);
	}

	if(outputflag == 0) {
		fprintf(stderr,"No output file specified\n");
		usage();
		exit(2);
	}

	if(perlout == 0 && bourneout == 0 && cshout == 0)
		perlout = 1;

	outfp = efopen(outputfile, "w");

	for(i=0; i<MAXFORMROWS; i++)
		answer[i][0] = '\0';
		
	/* readform fills in the editable, label, variable, and answer arrays */
	if((nrows = readform(filename)) == 0) {
		fprintf(stderr,"%s: empty form file.\n", progname);
		exit(1);
	}
	
	startcurses();
	signal(SIGQUIT, sighandle);
	signal(SIGINT, sighandle);

	/* initial update of the editing screen */
	for(i=0; i<nrows; i++) {
		move(i,0);
		addstr(label[i]);
		if(editable[i]) {
			addch(':');
			move(i, longval);	/* longval was determined by readform() */
			addstr(answer[i]);
		}
	}
	refresh();
	
	/* edit each of the editable lines in turn */
	i = 0;
	done = 0;
	while(1) {
		if(editable[i] == 1) {
			done = edit_line(i, longval, answer[i]);
			refresh();
			if(done == 1)		/* ESC was pressed */
				break;
		}
		if(done == 0) {			/* next line */
			if(++i == nrows)
				i = 0;
		} else if(done == -1) {	/* previous line */
			if(--i < 0)
				i = nrows - 1;
		}
	}

	/* output time */
	if(perlout)
		strcpy(fmt, PERLFMT);
	else if(bourneout)
		strcpy(fmt, BOURNEFMT);
	else if(cshout)
		strcpy(fmt, CSHFMT);
	else {		/* can't happen */
		fprintf(stderr,"neither perl, csh, nor bourne output specified.\n");
		exit(1);
	}

	for(i=0; i<nrows; i++) {
		if(editable[i] == 1)
			fprintf(outfp, fmt, variable[i], answer[i]);
	}

	finish_up();
	return(0);
}

/*
 - edit_line - perform edit operations on a single line
 *		The edited string is returned in the str argument.
 *		returns 1 if ESC typed, else 0 if RET typed
 */
edit_line(r, startc, str)
	int r;		/* row that we're editing */
	int startc;	/* starting column for editable string */
	char *str;	/* pointer to editable string */
{
	int k;
	int c;		/* current column */
	int endc;	/* ending column */
	int slen;	/* length of editable strings */
	int retcode;

	slen = strlen(str);
	c = endc = startc + slen;
	move(r, c);
	refresh();
	
	while(k = getch()) {
		switch(k) {
		case CTL('A'):
		case KEY_HOME:
			move(r, startc);
			c = startc;
			break;
		case CTL('F'):
		case KEY_RIGHT:
			c = ( c+1 <= endc ) ? c + 1 : c;
			move(r, c);
			break;
		case CTL('B'):
		case KEY_LEFT:
			c = ( c-1 >= startc ) ? c - 1 : c;
			move(r, c);
			break;
		case CTL('E'):
		case KEY_END:
			if(endc < COLS-2)
				c = endc;
			move(r, c);
			break;
		case CTL('D'):
			delch();
			endc--;
			break;
		case CTL('G'):
			help();
			move(r, c);
			break;
		case CTL('H'):
		case DEL:
		case KEY_BACKSPACE:
			if(c != startc) {
				c--;
				move(r, c);
				delch();
				endc--;
			}
			break;
		case CTL('K'):		/* delete to end of line */
			clrtoeol();
			endc = c;
			break;
		case CTL('L'):
			wrefresh(curscr);
			break;
		case CTL('I'):
		case CTL('N'):		/* next line */
		case '\n':
		case '\r':
		case KEY_DOWN:
			retcode = 0;
			goto finito;
		case CTL('P'):
		case KEY_UP:
			retcode = -1;
			goto finito;
		case ESC:			/* done with form */
			retcode = 1;
			goto finito;
		default:
			if(isprint(k)) {
				addch(k);
				if(c++ >= COLS-2) {
					c = COLS-2;
					move(r, c);
				}
				if(c > endc)
					endc++;
			} else {
				BEEP;
			}
			break;
		}
		refresh();
	}
finito:
	in_str(r, longval, endc, str);
	return(retcode);
}

/*
 - in_str - get the characters between startc and endc on a line
 */
in_str(row, startc, endc, str)
	int row;
	int startc;
	int endc;
	char *str;
{
	int i;
	char *cp;
	
	if(row < 0 || row > LINES - 1)
		return(-1);
	if(startc < 0 || startc > COLS - 1)
		return(-2);
	if(endc < 0 || endc > COLS - 1)
		return(-3);

	if(startc < endc) {
		for(cp=str,i=startc; i<endc; i++) {
			move(row, i);
			*cp++ = inch();
		}
		*cp = '\0';
	} else {		/* endc < startc */
		for(cp=str,i=startc; i>endc; i--) {
			move(row, i);
			*cp++ = inch();
		}
		*cp = '\0';
	}
	return(0);
}

/*
 - readform - read the lines from the formfile
 *
 * format of the form file:
 *		normal text lines - printed as they appear
 *		. lines - begin with a dot; editable lines
 *			.<label>~variable name~default value
 *
 *		example:
 *			.Name of Dog~dogname~Fido
 *			.Date of Vaccination~vacdate~9/17/91
 *
 * Read only up to MAXFORMROWS rows. Return the number of rows read.
 */
readform(file)
	char *file;
{
	char formrow[MAXFORMROWS][MAXSTR];	/* input form lines */
	FILE *formfp;	/* formfile file pointer */
	int i;
	int goteditable = 0;	/* non-zero if we get any editable lines */

	if(STREQ(file, "-"))
		formfp = stdin;
	else
		formfp = efopen(file, "r");

	for(i=0; i<MAXFORMROWS; i++) {
		if(fgets(formrow[i], MAXSTR, formfp) == NULL)
			break;
		CHOP(formrow[i], '\0');
		if(formrow[i][0] == '.') {
			goteditable++;
			editable[i] = 1;
			split(formrow[i]+1, FIELDSEP, 3, label[i], variable[i], answer[i]);
			if(strlen(label[i]) + 2 > longval)	/* find longest label */
				longval = strlen(label[i]) + 2;
		} else {
			strcpy(label[i], formrow[i]);
		}
	}
	fclose(formfp);
	if(formfp == stdin) {
		if(freopen("/dev/tty", "r+", stdin) != stdin) {
			fprintf(stderr,"%s: can't reopen stdin\n", progname);
			exit(2);
		}
	}

	if(goteditable == 0) {
		fprintf(stderr,"shmed: no editable lines in formfile! Quitting.\n");
		exit(3);
	}
	return(i);
}

/*
 - finish_up - clean up before exit
 */
void
finish_up()
{
	move(nrows, 0);
	refresh();
	nocbreak();
	echo();
	endwin();
}

/*
 - sighandle - catches SIGQUIT and SIGINT
 */
void
sighandle()
{
	signal(SIGQUIT, SIG_IGN);
	signal(SIGINT, SIG_IGN);
	finish_up();
	exit(1);
}

void
startcurses()
{
#ifdef NOTDEF
	if(initscr() == ERR) {
#else
	if(initscr() == NULL) {
#endif
		fprintf(stderr,"%s: can't start curses!\n", progname);
		exit(1);
	}
	noecho();
	cbreak();
	clear();

	altwin = newwin(LINES, COLS, 0, 0);
	helpwin = newwin(LINES, COLS, 0, 0);
	inithelp();
	(void)keypad(stdscr, TRUE);
	(void)keypad(helpwin, TRUE);
	(void)keypad(altwin, TRUE);
}

void
inithelp()
{
	mvwprintw(helpwin, 1, 12, "%s %s", progname, VERSION);
	mvwprintw(helpwin, 3, 0, "ctrl-N    go to next line (same as pressing RET)");
	mvwprintw(helpwin, 4, 0, "ctrl-P    go to previous line");
	mvwprintw(helpwin, 5, 0, "esc       form completed");

	mvwprintw(helpwin, 7, 0, "ctrl-A     go to beginning of line");
	mvwprintw(helpwin, 8, 0, "ctrl-E     go to end of line");
	mvwprintw(helpwin, 9, 0, "ctrl-F     go forward one space");
	mvwprintw(helpwin, 10, 0, "ctrl-B     go back one space");

	mvwprintw(helpwin, 12, 0, "ctrl-D     delete character at cursor");
	mvwprintw(helpwin, 13, 0, "backspace  delete character before cursor");
	mvwprintw(helpwin, 14,0, "ctrl-K     delete all characters to end of line");

	mvwprintw(helpwin, 16, 0, "ctrl-L    redraw screen");
	mvwprintw(helpwin, 17, 0, "ctrl-G    show help screen");
	mvwprintw(helpwin, 20, 10, "Hit any key to continue...");
}

void
help()
{
	overwrite(stdscr, altwin);
	overwrite(helpwin, stdscr);
	move(20, 36);
	refresh();
	wgetch(helpwin);
	overwrite(altwin, stdscr);
	refresh();
}
